/**
 * @file  gproxy/CommandLine.java
 **/
package gproxy;

import java.util.Hashtable;

/**
 * This class implements simple Command Line processing that is coded
 * over and over again in main methods. The current version is rather
 * simple.<p>
 *
 * The model is that command line arguments have the form:
 * <pre>  <option_args>* <free_form>*</pre>
 * where each element in <option_args> is an option starting
 * with a '-' and each element in <free_form> is a string.<p>
 *
 * By default options have the syntax:
 * <pre>  '-'NAME[=VALUE]</pre>
 * where NAME is the option identifier and VALUE is the string
 * argument for that option.<p>
 *<p>
 * Usage is as follows:
 * <pre>
 * static void main(String[] args) {
 *    CommandLine cl = new CommandLine();
 *    cl.parse(args);
 *    if (cl.getOption("verbose") != null) ...
 *    String file = cl.getArgument(0);
 *    String path = cl.getOption("classpath");
 * </pre>
 * @author Jan Vitek
 **/
public class CommandLine {
    private String[] commands_;
    private final Hashtable options_;
    private int offsetOfFirstArgument_;

    private static final int INITIAL_CAPACITY = 8;
    private static final char DASH = '-';
    private static final char EQUALS = '=';

    /**
     * Create a new CommandLine object.
     **/
    public CommandLine() {
	this.options_ = new Hashtable(INITIAL_CAPACITY);
    }

    /**
     * Create a new CommandLine object.
     * @param args parse these arguments
     **/
    public CommandLine(String[] args) {
	this();
	parse(args);
    }

    /**
     * Parse the command line arguments and extracts options.
     **/
    public void parse(String[] s) {
	offsetOfFirstArgument_ = 0;
	commands_ = s;

	for (int i = 0; i < commands_.length; i++) {
	    String str = commands_[i];
	    if ((str.length() > 0) && (str.charAt(0) == DASH)) {
		int eqPos = str.indexOf(EQUALS);
		String opt = null;
		String arg = "";
		if (eqPos > 0) {
		    arg = str.substring(eqPos + 1);
		    opt = str.substring(1, eqPos);
		} else 
		    opt = str.substring(1);
		options_.put(opt,arg);
		offsetOfFirstArgument_++;
	    } else {
		// We have seen all options. Bail out.
		return;
	    }
	}
    }

    /**
     * Return the String at offset i in the argument array or null if invalid.
     * @param i the offset in the argument array.
     * @return a string or null.
     **/
    public String getStringAt(int i) {
	if (i < 0 || i >= commands_.length) return null;
	return commands_[i];
    }

    /**
     * Return the ith argument String, options are not counted.
     * @param i the position
     * @return a string or null.
     **/
    public String getArgument(int i) {
	int p = i + offsetOfFirstArgument_;
	if (p < 0 || p >= commands_.length) return null;
	return commands_[p];
    }
    
    /**
     * Return the value of the option opt in the command line argument array.
     * @param opt an option String (without DASH)
     * @return the value of the option, or the empty String ("") if the
     * option has been set without a value, or null if the option has
     * not been set.
     **/
    public String getOption(String opt) {
	return (String) options_.get(opt);
    }

    /**
     * Return the number of non-option arguments.
     **/
    public int argumentCount() {
	return commands_.length - offsetOfFirstArgument_;
    }

    public CommandLine copyOptions() {
	CommandLine ret = new CommandLine();
	String[] args = new String[offsetOfFirstArgument_];
	System.arraycopy(commands_, 0, args, 0, offsetOfFirstArgument_);
	ret.parse(args);
	return ret;
    }

    public void addArgument(String newArg) {
	String[] args = new String[commands_.length + 1];
	System.arraycopy(commands_, 0, args, 0,  commands_.length);
	args[commands_.length] = newArg;
	parse(args);
    }

    public void addOption(String newArg) {
	String[] args = new String[commands_.length + 1];
	System.arraycopy(commands_, 0, args, 1,  commands_.length);
	args[0] = DASH + newArg;
	parse(args);
    }


} // End of CommandLine
 

