/**
 * Storage - management of the node's files.
 * @author Christian Grothoff
 * @file gnet/include/storage.h
 **/

#ifndef STORAGE_H
#define STORAGE_H

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <openssl/ripemd.h>
#include <wordexp.h>

#include "gnettypes.h"
#include "configuration.h"

/**
 * Build a filename from directory an fil parts.
 * @param fil the name of the file, will NOT be deallocated!
 * @returns the full file name, expanded with wordexp.
 **/
FileName expandFileName(FileName fil);

/**
 * implementation of "mkdir -p"
 * @param dir the directory to create
 **/
void mkdirp(DirName * dir);

/**
 * Build a filename from directory and filename, completing like the shell does
 * @param dir the name of the directory, may contain ~/ or other shell stuff. Will 
 *        NOT be freed!
 * @param fil the name of the file, will NOT be deallocated anymore!
 * @param result where to store the full file name (must be large enough!)
 **/
void buildFileName(DirName * dir,
		   HexName * fil,
		   FileName result);

/**
 * Read the contents of a binary file into a buffer.
 * @param fileName the name of the file, not freed,
 *        must already be expanded!
 * @param len the maximum number of bytes to read
 * @param result the buffer to write the result to
 * @return the number of bytes read on success, -1 on failure
 **/ 
int readFile(FileName fileName,
	     int  len,
	     void * result);

/**
 * Write a buffer to a file.
 * @param fileName the name of the file, NOT freed!
 * @param buffer the data to write
 * @param n number of bytes to write
 **/ 
void writeFile(FileName fileName, 
	       void * buffer,
	       int n);


/**
 * Write content to a file. Check for reduncancy and eventually
 * append.
 * @param fn the name of the file (without directory)
 * @param block the CONTENT_SIZE bytes long block (no header!)
 **/
void writeContent(HexName * fn, 
		  CONTENT_Block * block);

/**
 * Scan a directory for files. The name of the directory
 * must be expanded first (!).
 * @param dirName the name of the directory
 * @param result where to store the names of the files,
 *        only filenames with length 2*sizeof(HashCode160)
 *        will be stored.
 * @param max how many results should be obtained (max)
 * @return the number of files found, -1 on error
 **/
int scanDirectory(DirName * dirName,
		  DirContents * result,
		  int max);

/**
 * Convert (hash) block to hex (= filename)
 * @param block the sequence to convert
 * @param result where to store thestring (\0-terminated), hex-encoding
 **/
void hash2hex(HashCode160 * block,
	      HexName * result);

/**
 * Convert hex (filename) to the hostIdentity
 * @param hex the filename
 * @param ret address where to write the correspoinding host identity
 **/
void hex2hostIdentity(HexName * hex,
		      HostIdentity * ret);

/**
 * Convert hex (filename) to the hostIdentity
 * @param hex the filename
 * @param hash is set to the correspoinding host identity
 **/
void hex2hash(HexName * hex,
	      HashCode160 * hash);

/**
 * Hash block of given size.
 * @param block the data to hash, length is given as a second argument
 * @param ret pointer to where to write the hashcode
 **/
void hash(void * block,
	  int size,
	  HashCode160 * ret);
 
#endif

/* end of storage.h */
