/**
 * Main handler for incomming packets.
 * @author Christian Grothoff
 * @file gnet/node/handler.c
 **/

#include "handler.h"
#include "stdio.h"
#include "threads.h"
#include "keyservice.h"
#include "connection.h"

/**
 * The actual main method of gnet: message dispatch/handling.
 * And here is how it works:
 * 1) check integrity of the message (verify checksum)
 * 2) split message into submessages (eventually only one), and
 *    for each submessage, call the appropriate specific handler.
 *    If the message type is unknown, this part is skipped.
 * @param pack the message that was received. caller frees it on return
 * @param threadId number of the current thread
 **/
void handle(int threadId, MessagePack * pack) {
  BLOCK_LENGTH pos; /* position in the message */
  int veri; /* result from verify */
  MessagePart * part; /* part of the message */
  BLOCK_LENGTH total_len; /* length of message */
#ifdef PRINT_WARNINGS
  HexName hostName;
  IPPortString ipPort;
#endif

#ifdef DEBUG_PACKETS
  showMessageEncrypted(pack);
#endif
  /* verify packet, also decrypts! */  
  veri = verify(pack);
  if (veri<=0) {
#ifdef PRINT_WARNINGS
    if (veri == 0) {
      hash2hex(&pack->senderIdentity.hashPubKey, &hostName);
      printAddress(&pack->sender, &ipPort);
      print("CRC error.\n Dropping packet from %s-%s.\n",
	    &hostName, &ipPort);
    }
#endif
    return; /* CRC error, abort */
  }
#ifdef DEBUG_PACKETS
  showMessageDecrypted(pack);
#endif  
  total_len = pack->size+BLOCK_LENGTH_SIZE - sizeof(MessagePack);
  pos = 0;
  /* split into submessages, call subhandlers */
  if (veri == 2) {  /* message was not encrypted! */
    while (pos < total_len) {
      part = (MessagePart*) &((unsigned char*)&pack->parts[0])[pos];
      pos += part->size+BLOCK_LENGTH_SIZE;
      if ((pos > total_len)) {
#ifdef PRINT_WARNINGS
	print("Header size is %d\n.",sizeof(MessagePack));
	print("Invalid packet format (%d - %d)- dropping remaining parts (%d).\n",
	      pos, total_len,
	      part->size);
#endif
	print("Packet contained (%d):\n",veri);
	dumpPacket((unsigned char*) &pack->parts[0], total_len);
	return;
      }      
#ifdef PRINT_UDP
      print("UDP: Handling message part (unencrypted) with size %d and type %d.\n",
	    part->size,part->requestType);
#endif
      switch (part->requestType) { 	
      case GNET_PROTO_HELO:
	handleHELO((HELO_Message*)part); 
	break;
      case GNET_PROTO_SKEY:  
	handleSKEY(pack,(SKEY_Message*)part);
	break;
      /* This would be the place to add more handlers for NON-encrypted 
	 messages (none planned yet, though) */
      default:
#ifdef PRINT_WARNINGS
	print("Unknown subprotocol (NOT encrypted): %d. Part ignored.\n",
	      part->requestType);
#endif
	break;
      }    
    } /* for all parts */
  } else {
    /* message was encrypted */
    while (pos < total_len) {
      part = (MessagePart*) &((unsigned char*)&pack->parts[0])[pos];
      pos += part->size+BLOCK_LENGTH_SIZE;
      if (pos > total_len) {
#ifdef PRINT_WARNINGS
	print("Invalid e-packet format (%d - %d) - dropping remaining parts.\n",
	      pos, total_len);
	print("Packet contained (%d):\n",veri);
	dumpPacket((unsigned char*)&pack->parts[0], total_len);
#endif
	return;
      }      
      switch (part->requestType) {		
      case GNET_PROTO_HELO:
	handleHELO((HELO_Message*)part); 
	break;
      case GNET_PROTO_SKEY:
	handleSKEY(pack,(SKEY_Message*)part);
	break;
      case GNET_PROTO_QUERY:
	notifyActive(&pack->senderIdentity);
	handleQUERY(pack, (QUERY_Message*)part);
	break;
      case GNET_PROTO_SKEY_ACK:
	handleSKEYACK(pack,(ACK_Message*)part);
	break;
      case GNET_PROTO_CONTENT:
	notifyActive(&pack->senderIdentity);
	handleCONTENT(pack, (CONTENT_Message*)part);
	break;       
      case GNET_PROTO_3QUERY:
	notifyActive(&pack->senderIdentity);
	handle3QUERY(pack, (QUERY_Message*)part);
	break;
       
      /* This would be the place to add more handlers for ENCRYPTED messages 
	 (planned: status information, credit trading, routing?) */
      default:
#ifdef PRINT_WARNINGS
	print("Unknown subprotocol (encrypted): %d - part ignored!\n",
	      part->requestType);
#endif
	break;
      }    
    } /* for all parts */
  } /* end of if-encrypted */
} /* end of handle */

/**
 * Verify that this message is correct (CRC, sym. encoding).
 * How it works:
 * 0) check policy: do we currently accept anything from that source?
 * 1) IF highest bit in size is set, decrypt:
 *    decrypt packet with key from local database
 *    (senderPort + senderIP => symmetric key)
 *    --- if not encrypted: return 2 at the end!
 *    Important: also the checksum is encoded!
 * 2) check CRC checksum for decrypted packet (rest!)
 *    If wrong, return 0.
 * @param pack the message that was received and that should be checked
 **/
int verify(MessagePack * pack) {
  int returnValue = 2;
  unsigned char decr[MESSAGE_SIZE];
  int size;
#ifdef PRINT_WARNINGS
  HexName hostName;
#endif
  
  /* check policy (is sender host blocked? is load ok?) */
  if (!incomingCheck(&pack->senderIdentity)) {
    return -1; /* error: packet denied by policy */
  }

  /* decrypt packet */
  if ((pack->size & ENCRYPTED_FLAG) > 0) {          
    pack->size = pack->size & (ENCRYPTED_FLAG-1); /* mask out highest bit */
    /* decrypt msg */
    size = decryptFromHost((BLOWFISHEncryptedData*)&pack->parts[0],
			   pack->size - sizeof(MessagePack) + BLOCK_LENGTH_SIZE,
			   &pack->senderIdentity,
			   decr);
    if (-1 == size) {
#ifdef PRINT_WARNINGS
      hash2hex(&pack->senderIdentity.hashPubKey, &hostName);
      print("Can not decrypt packet from host %s: no session key! Returning CRC error.\n",
	    &hostName);
#endif
      return 0;
    }
    memcpy(&pack->parts[0],
	   decr, size);
    returnValue = 1;
  } else
    size = pack->size - sizeof(MessagePack) + BLOCK_LENGTH_SIZE;
  
  /* Check CRC vs. information in header */
#ifndef DISABLE_CRC_CHECK
  if (pack->checksum != crc32N(&pack->parts[0],size)) {
    /* invalid CRC */
#ifdef PRINT_CRC 
    print("CRC failed. Expected %d, got %d.\n",
	  pack->checksum, 
	  crc32N(&pack->parts[0], size));	  
#endif 
#ifdef PRINT_WARNINGS
    print("CRC ERROR: %d (pack) %d (msg)\n", 
	  pack->checksum, 
	  crc32N(&pack->parts[0],size));
#endif
    return 0; 
  }
#ifdef PRINT_CRC 
  print("CRC ok.\n");
#endif 
#endif
  return returnValue; 
}


/* ******************* Methods to handle the different requests ************* */

/**
 * Take the HELO message and add the public key to the
 * list of known hosts. Also bind the identity to the
 * connection information provided.
 **/
void handleHELO(HELO_Message * msg) {
#ifdef PRINT_UDP
  IPPortString ipPort;

  printAddress(&msg->body.sender,&ipPort);
  print("UDP: Handling HELO from %s.\n",
	&ipPort);    
#endif
  if (1 == verifyHELO(msg)) {
#ifdef PRINT_WARNINGS
    print("HELO message received invalid. Dropping.\n");
#endif
    return; /* message invalid */  
  }
  bindAddress(msg); 

#ifdef DEBUG_HELO_TRIGGERS_QUERY
  DEBUG_broadcastQuery();
#endif
#ifdef PRINT_UDP_ALL
    print("UDP: Done handling HELO\n");    
#endif
}

/**
 * Signed session key exchange.
 **/
void handleSKEY(MessagePack * pack, SKEY_Message * msg) {
#ifdef PRINT_UDP
  IPPortString ipPort;

  printAddress(&pack->sender,&ipPort);
  print("UDP: Handling SKEY from %s.\n",
	&ipPort);    
#endif  
  acceptSessionKey((HostIdentity *) &pack->senderIdentity, 
		   msg);  
#ifdef PRINT_UDP_ALL
  print("UDP: Done handling SKEY\n");    
#endif
}

/**
 * Signed session key exchange.
 **/
void handleSKEYACK(MessagePack * pack, ACK_Message * msg) {
#ifdef PRINT_UDP
  IPPortString ipPort;

  printAddress(&pack->sender,&ipPort);
  print("UDP: Handling SKEY ACK from %s.\n",
	&ipPort);    
#endif

  notifyActiveSKEY(&pack->senderIdentity, 
		   &msg->hash);
#ifdef PRINT_UDP
    print("UDP: Done handling SKEY ACK.\n");    
#endif
}

/**
 * Handle query for content. Depending on how we like
 * the sender, lookup or even forward.
 **/
void handleQUERY(MessagePack * pack, QUERY_Message * msg) {
  QUERY_POLICY qp;
#ifdef PRINT_UDP
  IPPortString ipPort;
  HexName query;
#endif
#ifdef PRINT_UDP_ALL
#ifndef PRINT_UDP
  HexName query;
#endif
#endif

  if (msg->ttl <= 0)
    return;

#ifdef PRINT_UDP
  hash2hex(&msg->query, &query);
  printAddress(&pack->sender,&ipPort);
  print("UDP: Handling QUERY for %s from %s.\n",
	&query, &ipPort);
#endif
 /* first: query policy */
  qp = evaluateQuery(&pack->senderIdentity,
		     msg->priority);  
  if ((qp & QUERY_DROPMASK) == 0)
    return; /* straight drop. */
  /* if ok: do query */
  execQuery(qp, msg);
#ifdef PRINT_UDP_ALL
#ifndef PRINT_UDP
  hash2hex(&msg->query, &query);
#endif
  print("UDP: Done handling QUERY %s.\n",
	&query);    
#endif
}

/**
 * Handle query for content. Depending on how we like
 * the sender, lookup or even forward.
 **/
void handle3QUERY(MessagePack * pack, QUERY_Message * msg) {
  QUERY_POLICY qp;
#ifdef PRINT_UDP
  IPPortString ipPort;
  HexName query;
#endif
#ifdef PRINT_UDP_ALL
#ifndef PRINT_UDP
  HexName query;
#endif
#endif

  if (msg->ttl <= 0)
    return;

#ifdef PRINT_UDP
  hash2hex(&msg->query, &query);
  printAddress(&pack->sender,&ipPort);
  print("UDP: Handling 3QUERY for %s from %s.\n",
	&query, &ipPort);
#endif
 /* first: query policy */
  qp = evaluateQuery(&pack->senderIdentity,
		     msg->priority);  
  if ((qp & QUERY_DROPMASK) == 0)
    return; /* straight drop. */
  /* if ok: do query */
  exec3Query(qp, msg);
#ifdef PRINT_UDP_ALL
#ifndef PRINT_UDP
  hash2hex(&msg->query, &query);
#endif
  print("UDP: Done handling 3QUERY %s.\n",
	&query);    
#endif
}

/**
 * Receive content, do something with it!
 * There are 3 basic possiblilities. Either our
 * node did the request and we should send the result
 * to gproxy via TCP, or the content was requested by
 * another node and we forwarded the request (and thus
 * we now have to fwd the reply) or 3rd somebody just
 * send us some content we did NOT ask for - and we
 * can choose to store it or just discard it.
 * <p>
 * handleCONTENT leaves the possibilities gproxy and
 * indirection to "useContent". The policy module
 * is considered for local storage.
 **/
void handleCONTENT(MessagePack * pack, 
		   CONTENT_Message * msg) {  
  int prio;

#ifdef PRINT_UDP
  HexName cont;

  hash2hex(&msg->reply.hash, &cont);
  print("UDP: Handling CONTENT %s.\n",
	&cont);    
#endif

  if (msg->header.size != 
      (sizeof(CONTENT_Message)-BLOCK_LENGTH_SIZE)) {
#ifdef PRINT_WARNINGS
    print("Received invalid content packet (size %d), dropping.\n",
	  msg->header.size);
#endif  
    return; 
  }
  writeBack(&msg->reply); /* this sends it to GProxy */
  prio = useContent(&pack->senderIdentity,
		    &msg->reply.content, 
		    &msg->reply.hash);
  if (pack != NULL) /* never evaulate loopback-content! */
    evaluateContent(&pack->senderIdentity,
		    &msg->reply.content,
		    &msg->reply.hash, 
		    prio);
#ifdef PRINT_UDP_ALL
  print("UDP: Done handling CONTENT.\n");    
#endif
}


/* end of handler.c */
