/**
 * The server side of the TCP connection between GProxy and Gnet node
 * @file gnet/node/tcpListener.c
 * @author Tiberiu Stef
 * @author Christian Grothoff
 **/

#include <stdio.h>
#include <stdlib.h> 	 
#include <unistd.h>		
#include <pthread.h>		
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h> 

#include "tcpListener.h"
#include "configuration.h"
#include "gnettypes.h"
#include "policy.h"
#include "keyservice.h"
#include "semaphore.h"

/**
 * Array of currently open TCP sockets to GProxies.
 **/
int * TCPLISTENER_sockDescriptors_;

/**
 * We must guard concurrent access to the TCPLISTENER_sockDescriptors_ array
 **/
Semaphore * TCPLISTENER_sockDescriptorGuard_;

/**
 * The thread that waits for new connections.
 **/
pthread_t TCPLISTENER_listener_;

/**
 * Handle data available on the TCP socket descriptor :
 * it's a simple echo, will return back everything received
 * @param sockDescriptor the socket that we are listening to (fresh)
 **/
void processData(int sockDescriptor){
  GPROXY_Message buff;
  int n;
  QUERY_POLICY qp = QUERY_ANSWER + QUERY_FORWARD + QUERY_PRIORITY_BITMASK; 
  QUERY_Message msg;
  HostIdentity me;
#ifdef PRINT_TCP
  HexName hex;
#endif  
  semaphore_down(TCPLISTENER_sockDescriptorGuard_);
  for (n=0;n<CONCOUNT;n++) 
    if (TCPLISTENER_sockDescriptors_[n] == -1) {
      TCPLISTENER_sockDescriptors_[n] = sockDescriptor;
      break;
    }
  semaphore_up(TCPLISTENER_sockDescriptorGuard_);
#ifdef DEBUG_TCP
  print("TCP: incoming TCP connection.\n");
#endif

  if (n == CONCOUNT) {
#ifdef PRINT_WARNINGS
    print("Maximum number of TCP GProxy Clients reached (%d)!\n",
	  n);
#endif
    return;
  }
 
#ifdef PRINT_TCP
  print("TCP: opened connection \n");
#endif
  
  getHostIdentity(getPublicHostkey(),&me);
  n = 1;
  while (n > 0){
    n = read(sockDescriptor,&buff,sizeof(GPROXY_Message));
    if (n != sizeof(GPROXY_Message)) 
      break;
    /* convert double-hash to triple-hash */
#ifdef PRINT_TCP
    hash2hex(&buff.query, &hex);
    print("TCP: received double-query %s.\n",
	  &hex);
#endif
    hash(&buff.query,
	 sizeof(HashCode160),
	 &msg.query);
    msg.priority = buff.priority;
    msg.ttl = buff.ttl;
    memcpy(&msg.returnTo, &me, sizeof(HostIdentity));
    msg.header.size = sizeof(QUERY_Message)-BLOCK_LENGTH_SIZE;
    msg.header.requestType = GNET_PROTO_3QUERY;
#ifdef PRINT_TCP
    hash2hex(&msg.query, &hex);
    print("TCP: executing triple-query %s.\n",
	  &hex);
#endif
    tripleHashLookup(&msg.query,
		     &msg.returnTo,
		     qp);
    indirectingQuery(&msg); 
    forwardQuery(&msg);
    
#ifdef PRINT_TCP_ALL
    print("TCP: query sent.\n");
#endif
  }
#ifdef PRINT_TCP
  print("TCP: closing connection.\n");
#endif

  semaphore_down(TCPLISTENER_sockDescriptorGuard_);
  for (n=0;n<CONCOUNT;n++) 
    if (TCPLISTENER_sockDescriptors_[n] == sockDescriptor) {
      TCPLISTENER_sockDescriptors_[n] = -1;
      break;
    }
  semaphore_up(TCPLISTENER_sockDescriptorGuard_);
}

/**
 * Method called whenever a new thread is created, when there is a new 
 * connection to the node. Main method of the thread. 
 * Will echo the data sent and then exit
 * @param descriptor the socket to listen to
 **/
void incomingProcessor(int descriptor){ 
#ifdef PRINT_TCP_ALL
  print("TCP: Processing thread for descriptor %d\n",descriptor);
#endif
  processData(descriptor);
#ifdef PRINT_TCP_ALL
  print("TCP: Terminating thread %d...\n",descriptor);
#endif
  close(descriptor);
  pthread_exit(NULL);
}

/**
 * Writes the data given in buff to all current GProxy connections.
 * @param buff the data to send to the GProxies.
 **/
void writeBack(GPROXY_Reply * buff) {
  int i;
  int sd;
  /* send the data back - echo */  
#ifdef PRINT_TCP
  HexName hex;

  hash2hex(&buff->hash, &hex);
  print("TCP: begin sending %d bytes to gproxy matching query %s.\n",
	sizeof(GPROXY_Reply),
	&hex);
#endif
  for (i=0;i<CONCOUNT;i++) {
    sd = TCPLISTENER_sockDescriptors_[i]; /* avoid too many syncs... */
    if (sd != 0) 
      /* WAS: write(TCPLISTENER_sockDescriptors_[i],&buff[BLOCK_LENGTH_SIZE],len(buff));  */
      /* if we'd use write, we must catch SIGPIPE signal! 
         but async send is better anyway */
      send(TCPLISTENER_sockDescriptors_[i],
	   buff,
	   sizeof(GPROXY_Reply),
	   MSG_DONTWAIT|MSG_NOSIGNAL); /* send message, don"t bother if errors occur */
    /* the sd may be closed here, but we would not crash in that case! */
  }
#ifdef PRINT_TCP
  print("TCP: done sending content to gproxy.\n");
#endif
}

/**
 * Initialize the TCP port and listen for incomming GProxy connections.
 **/
void tcpListenMain() {
  int listenerFD,incomingFD;
  int lenOfIncomingAddr;
  int listenerPort=getGnetPort();
  struct sockaddr_in serverAddr, clientAddr;
  pthread_t incomingThread; 	/* thread handling a new connection */
  int x;
  int secs = 5;

  TCPLISTENER_sockDescriptorGuard_ = new_semaphore(1);
  TCPLISTENER_sockDescriptors_ = (int*) malloc(sizeof(int)*CONCOUNT);
  for (x=0;x<CONCOUNT;x++)
    TCPLISTENER_sockDescriptors_[x] = -1;

  /* create the socket */
  while ( (listenerFD=socket(PF_INET,SOCK_STREAM,0)) < 0) {
#ifdef PRINT_WARNINGS
    print("Error opening socket. No proxy service started. Trying again in 30 seconds.\n");
#endif
    sleep(30);
  }
  
  /* fill in the inet address structure */
  bzero((char *) &serverAddr, sizeof(serverAddr));
  serverAddr.sin_family = AF_INET;
  serverAddr.sin_addr.s_addr=htonl(INADDR_ANY);
  serverAddr.sin_port=htons(listenerPort);
  
  /*bind the socket */
  while (bind (listenerFD, (struct sockaddr *) &serverAddr,sizeof(serverAddr)) < 0){
#ifdef PRINT_WARNINGS
    print("Error binding the TCP listener to port %d. No proxy service started.\nTrying again in %d seconds...\n",
	  listenerPort,secs);
#endif
    sleep(secs);
    secs+=5; /* slow progression... */
  }
  
  /* start listening for new connections */
  listen(listenerFD,CONCOUNT); /* max: 5 pending connections */
  
  /*process incoming data */
  while (1) {
    /*wait for a connection and process it */
    lenOfIncomingAddr=sizeof(clientAddr);
    incomingFD=accept(listenerFD,(struct sockaddr *)&clientAddr, &lenOfIncomingAddr);
    if(incomingFD < 0){
#ifdef PRINT_WARNINGS
      print("Error accepting new connection.\n");
#endif
      continue;
    }
#ifdef PRINT_TCP
    print("TCP: starting GProxy server\n");
#endif
    if ((pthread_create(&incomingThread, NULL,
			(void * (*)(void *)) incomingProcessor, (void *) incomingFD)) != 0){
#ifdef PRINT_WARNINGS
      print("Error creating thread to handle new incoming connection.\n");    
#endif
      close(incomingFD);
    }
  } /* while (1) */
} 

/**
 * Initialize the TCP port and listen for incomming GProxy connections.
 **/
void startTCPListener() {
  pthread_create(&TCPLISTENER_listener_, NULL,
		 (void * (*)()) &tcpListenMain, NULL); 
}


/* end of tcpListener */
