/**
 * Configuration management.
 * @author Christian Grothoff
 * @file primitives/configuration.c
 **/

#include "configuration.h"

#include <netdb.h>
#include <stdio.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "storage.h"


#define LIMITS_SECTION        "LIMITS"
#define DISK_QUOTA            "DISKQUOTA"
#define MAX_LOAD              "MAXLOAD"
#define MAX_NET               "MAXNET"
#define MAX_NET_LOAD          "MAXNETLOAD"
#define MAX_MEM               "MEMQUOTA"
#define MAX_NODES             "MAXNODES"
#define GNET_PORT             "PORT"
#define GNET_SECTION          "GNET"
#define HOST_KEY_FILE_PRIV    "HOSTKEY"
#define DATABASE_FILE         "DATABASE"
#define COLLISION_FILE	      "COLLISION"
#define SHARED_FILE_LIST      "FILELIST"
#define CONTENT_DIRECTORY     "CONTENTDIR"
#define CREDIT_DIRECTORY      "CREDITDIR"
#define HOST_KEY_DIRECTORY    "HOSTKEYS"
#define HELO_EXPIRES          "HELOEXPIRES"
#define NETWORK_IDENTITIES    "IDENTITIES"
#define NETWORK_INTERFACES    "INTERFACES"
#define NETWORK_IP            "IP"


/**
 * PRIVATE: gnet configuration
 **/
LHASH * config_;

void readConfig(FileName filename) {
  long errorLine; /* in which line did an error occur */
  FileName fn;
  
  fn = expandFileName(filename);
#ifdef PRINT_CONFIGURATION
  print("Loading configuration from %s.\n", fn);
#endif  
  config_ = CONF_load(NULL, fn, &errorLine);
  free(fn);
  if (config_ == NULL) {
    ERR_load_crypto_strings();
    print("Unable to load configuration from %s. Error at line %ld.\n",
	  filename,
	  errorLine);
    ERR_print_errors_fp(stderr);
    exit(1);
  }
  /* create directories */
  mkdirp(getContentDirectory());
  mkdirp(getNetworkIdDirectory());
  mkdirp(getCreditDirectory());
}

void freeConfig() {
#ifdef PRINT_CONFIGURATION
  print("Freeing configuration.\n");
#endif  
  CONF_free(config_);
}

unsigned int getDiskQuota() {
  return (unsigned int) CONF_get_number(config_, 
				      LIMITS_SECTION,
				      DISK_QUOTA);
}

unsigned int getMaxLoad() {
  return (unsigned int) CONF_get_number(config_, 
				      LIMITS_SECTION,
				      MAX_LOAD);
}

unsigned int getMaxNetLoad() {
  return (unsigned int) CONF_get_number(config_, 
				      LIMITS_SECTION,
				      MAX_NET_LOAD);
}

unsigned int getMaxNet() {
  return (unsigned int) CONF_get_number(config_, 
				      LIMITS_SECTION,
				      MAX_NET);
}

IPAddress getIP() {
  IPAddress i = CONF_get_string(config_,
				GNET_SECTION,
				NETWORK_IP);
  return i;
}

Interfaces getNetInterfaces() {
  Interfaces i = CONF_get_string(config_,
				 GNET_SECTION,
				 NETWORK_INTERFACES);
  return i;
}

unsigned int getMaxMem() {
  return (unsigned int) CONF_get_number(config_, 
				      LIMITS_SECTION,
				      MAX_MEM);
}
unsigned int getMaxNodes() {
  return (unsigned int) CONF_get_number(config_, 
				      LIMITS_SECTION,
				      MAX_NODES);
}

/**
 * How many hours is a HELO message
 * valid?
 **/
unsigned int getHeloExpiration() {
  unsigned int exp = (unsigned int) CONF_get_number(config_, 
						    GNET_SECTION,
						    HELO_EXPIRES);
  if (exp < 1) 
    exp = 1;
  return exp;
}

unsigned short getGnetPort() {
  struct servent	*pse;	/* pointer to service information entry	*/
  unsigned short port 
    = (unsigned short) CONF_get_number(config_, 
				       GNET_SECTION,
				       GNET_PORT);
  if (port == 0) { /* try lookup in services */
    if ((pse = getservbyname("gnet", "udp"))) 
      port = htons(pse->s_port);      
    else 
      errexit("Cannot determine port to bind to. Define in configuration file in section %s under %s or in /etc/services under UDP/gnet.\n",
	      GNET_SECTION, GNET_PORT);
  }
  return port;
}

FileName getHostKeyFilePriv() {
  FileName f = CONF_get_string(config_,
			       GNET_SECTION,
			       HOST_KEY_FILE_PRIV);
  if (f == NULL) 
    errexit("Configuration file must specify filename for hostkey in section %s under %s.\n",
	    GNET_SECTION,
	    HOST_KEY_FILE_PRIV);
  return f;
}

FileName getDatabaseFile() {
  FileName f = CONF_get_string(config_,
			       GNET_SECTION,
			       DATABASE_FILE);
  if (f == NULL) 
    errexit("Configuration file must specify filename for the database in section %s under %s.\n",
	    GNET_SECTION,
	    DATABASE_FILE);
  return f;
}

FileName getCollisionFile() {
  FileName f = CONF_get_string(config_,
                               GNET_SECTION,
                               COLLISION_FILE);
  if (f == NULL)
     errexit("Configuration file must specify filename for collisions in section %s under %s.\n",
            GNET_SECTION,
            COLLISION_FILE);
  return f;
}

FileName getSharedFileList() {
  FileName f = CONF_get_string(config_,
			       GNET_SECTION,
			       SHARED_FILE_LIST);
  if (f == NULL) 
    errexit("Configuration file must specify filename for the list of on-demand shared files in section %s under %s.\n",
	    GNET_SECTION,
	    SHARED_FILE_LIST);
  return f;
}

DirName * getCreditDirectory() {
  DirName * f 
    = (DirName*) CONF_get_string(config_,
				GNET_SECTION,
				CREDIT_DIRECTORY);
  if (f == NULL) 
    errexit("Configuration file must specify directory for credits in section %s under %s.\n",
	    GNET_SECTION,
	    CREDIT_DIRECTORY);
  return f;
}

DirName * getContentDirectory() {
  DirName * f 
    = (DirName*) CONF_get_string(config_,
				 GNET_SECTION,
				 CONTENT_DIRECTORY);
  if (f == NULL) 
    errexit("Configuration file must specify directory for content in section %s under %s.\n",
	    GNET_SECTION,
	    CONTENT_DIRECTORY);
  return f;
}

DirName * getNetworkIdDirectory() {
  DirName * f = (DirName*) CONF_get_string(config_,
					   GNET_SECTION,
					   NETWORK_IDENTITIES);
  if (f == NULL) 
    errexit("Configuration file must specify directory for network identities in section %s under %s.\n",
	    GNET_SECTION,
	    NETWORK_IDENTITIES);
  return f;
}



/* end of configuration.c */
