/**
 * Module for cron jobs.
 * @author Christian Grothoff
 * @file src/primitives/cron.c
 **/

#include "cron.h"

/**
 * The delta-list of waiting tasks.
 **/
DeltaEntry * CONTROLTHREAD_deltaList_;

/** 
 * The lock for the delta-list.
 **/
Semaphore  * CONTROLTHREAD_deltaListLock_;

/**
 * The first empty slot in the delta-list.
 **/
int CONTROLTHREAD_firstFree_;

/**
 * The first empty slot in the delta-list.
 **/
int CONTROLTHREAD_firstUsed_;

/**
 * The cron thread.
 **/
pthread_t CONTROLTHREAD_cron_;

/**
 * Initialize the cron service.
 **/
void initCron() {
  int i;

  CONTROLTHREAD_deltaList_ 
    = malloc(sizeof(DeltaEntry) * MAX_CRON_JOBS);
  for (i=0;i<MAX_CRON_JOBS;i++)
    CONTROLTHREAD_deltaList_[i].next = i-1;
  CONTROLTHREAD_firstFree_ = MAX_CRON_JOBS-1;
  CONTROLTHREAD_deltaListLock_
    = new_semaphore(1);
  CONTROLTHREAD_firstUsed_
    = -1;
}

/**
 * Start the cron jobs.
 **/
void startCron() {

#ifndef DISABLE_CONTROL_THREAD
#ifdef PRINT_THREADS
  print("THREADS: Creating control thread.\n");
#endif
  pthread_create(&CONTROLTHREAD_cron_, NULL,
		 (void * (*)()) cron, NULL);
#else
  print("WARNING: Control Thread Not Created!\n");
#endif
}

/**
 * Add a cron-job to the delta list.
 * @param method which method should we run
 * @param delta how many seconds until we run the method
 * @param deltaRepeat if this is a periodic, the time between
 *        the runs, otherwise 0.
 **/
void addCronJob(void (*method)(),
		int delta,
		int deltaRepeat) {
  DeltaEntry * entry;
  DeltaEntry * pos;
  int last;
  int current;

  semaphore_down(CONTROLTHREAD_deltaListLock_);
  if (CONTROLTHREAD_firstFree_ == -1) {
    semaphore_up(CONTROLTHREAD_deltaListLock_); 
#ifdef PRINT_WARNINGS
    print("CONTROLTHREAD: No more slots for cron jobs!\n");
    return;
#endif    
  }
  entry = &CONTROLTHREAD_deltaList_[CONTROLTHREAD_firstFree_];
  entry->method = method;
  entry->deltaRepeat = deltaRepeat;
  if (CONTROLTHREAD_firstUsed_ == -1) {
    CONTROLTHREAD_firstUsed_ 
      = CONTROLTHREAD_firstFree_;
    CONTROLTHREAD_firstFree_
      = entry->next;
    entry->next = -1; /* end of list */
    entry->delta = delta;
    semaphore_up(CONTROLTHREAD_deltaListLock_); 
    return;
  }
  /* no, there are jobs waiting */
  last = -1;
  current = CONTROLTHREAD_firstUsed_;
  pos = &CONTROLTHREAD_deltaList_[current];
  while (delta > pos->delta) {
    delta -= pos->delta;
    if (pos->next != -1) {
      last = current;
      current = pos->next;
      pos = &CONTROLTHREAD_deltaList_[current];
    }
    else { /* append */
      pos->next = CONTROLTHREAD_firstFree_;
      entry->delta = delta;
      CONTROLTHREAD_firstFree_
	= entry->next;
      entry->next = -1;
      semaphore_up(CONTROLTHREAD_deltaListLock_); 
      return;
    }      
  }
  /* insert before pos */
  if (last == -1)
    CONTROLTHREAD_firstUsed_ = CONTROLTHREAD_firstFree_;
  else 
    CONTROLTHREAD_deltaList_[last].next = CONTROLTHREAD_firstFree_;
  entry->delta = delta;
  pos->delta -= delta;
  CONTROLTHREAD_firstFree_
    = entry->next;
  entry->next = current;
  semaphore_up(CONTROLTHREAD_deltaListLock_); 
}

/**
 * Print the cron-tab.
 **/
void printCronTab() {
  int jobId;
  DeltaEntry * tab;

  semaphore_down(CONTROLTHREAD_deltaListLock_);
  print("Start: %d Free: %d\n", 
	CONTROLTHREAD_firstUsed_,CONTROLTHREAD_firstFree_);
  for (jobId = 0;jobId<MAX_CRON_JOBS;jobId++) {
    tab = &CONTROLTHREAD_deltaList_[jobId];
    print("%d: delta %d --- method %d --- repeat %d --- next %d\n",
	  jobId, tab->delta, tab->method, tab->deltaRepeat, 
	  tab->next);
  }
  semaphore_up(CONTROLTHREAD_deltaListLock_);
}

/**
 * Process the cron-job at the beginning of the waiting
 * queue, that is, remove, invoke, and re-insert if
 * it is a periodical job. Make sure the sync is down
 * while the job is running (it may add other jobs!)
 **/
void runJob() {
  DeltaEntry * job;
  int jobId;
  void (* method)();
  int repeat;

  semaphore_down(CONTROLTHREAD_deltaListLock_);
  jobId = CONTROLTHREAD_firstUsed_;
  if (jobId == -1) {
    semaphore_up(CONTROLTHREAD_deltaListLock_); 
    return; /* no job to be done */
  }
  job = &CONTROLTHREAD_deltaList_[jobId];
  method = job->method;
  repeat = job->deltaRepeat;
  /* remove from queue */
  CONTROLTHREAD_firstUsed_
    = job->next;
  job->next
    = CONTROLTHREAD_firstFree_;
  CONTROLTHREAD_firstFree_ = jobId;
  semaphore_up(CONTROLTHREAD_deltaListLock_); 
  /* run */
  method();
  /* re-insert */
  if (repeat > 0) 
    addCronJob(method, repeat, repeat);
}

/**
 * The main-method of cron.
 **/
void cron() {
  int delta;
  while (1) {
    semaphore_down(CONTROLTHREAD_deltaListLock_);
    if (CONTROLTHREAD_firstUsed_ != -1) 
      CONTROLTHREAD_deltaList_[CONTROLTHREAD_firstUsed_].delta
	-= CRON_GRANULARITY;
    delta = -1;
    while (delta <= 0) {
      if (delta == 0)
	semaphore_down(CONTROLTHREAD_deltaListLock_);
      if (CONTROLTHREAD_firstUsed_ != -1) {
	delta = CONTROLTHREAD_deltaList_[CONTROLTHREAD_firstUsed_].delta; 
	if (delta < 0) {
	  print("WARNING: crontab inconsistend, cron exits!\n");
	  semaphore_up(CONTROLTHREAD_deltaListLock_);     
	  return;
	}
      } else
	delta = 1; /* end loop */
      semaphore_up(CONTROLTHREAD_deltaListLock_);     
      if (delta == 0)
	runJob();
    }
    sleep(CRON_GRANULARITY); /* granularity here! */
  }
}

/* end of cron.c */
