/**
 * GNet Debugging module
 * @author Christian Grothoff
 * @file primitives/debugging.c
 **/

#include "gnettypes.h"
#include "sessionkey.h"

/**
 * print a debug message
 * @param format the string describing the error message
 **/
void print(const char *format, ...) {
  va_list	args;  
  va_start(args, format);
  vfprintf(PRINTTO, format, args);
  va_end(args);
}

/**
 * errexit - print an error message and exit
 * @param format the string describing the error message
 **/
void errexit(const char *format, ...) {
  va_list args;
  va_start(args, format);
  vfprintf(stderr, format, args);
  va_end(args);
  exit(1);
}

/**
 * Display the contents of an incomming UDP message.
 **/
void showMessageEncrypted(MessagePack * pack) {
  IPPortString ipPort;

  printAddress(&pack->sender,
	       &ipPort);
  print("Incomming packet from %s (size %5d-%4d, CRC %5d)\n",
	&ipPort,
	pack->size, pack->size & 0x7FFF,
	pack->checksum);
}

/**
 * Display the contents of an incomming UDP message.
 **/
void showMessageDecrypted(MessagePack * pack) {
  IPPortString ipPort;

  printAddress(&pack->sender,
	       &ipPort);
  print("Decrypted packet from %s (size %4d-%5d, CRC %5d)\n",
	&ipPort,
	pack->size, pack->size & 0x7FFF,
	pack->checksum);
}

/**
 * Print the connection address (IP and Port) encoded in the given
 * HostAddress to stderr.
 * @param conInfo the HostAddress to print.
 **/
void printAddress(HostAddress * conInfo,
		  IPPortString * result) {
  if (conInfo == NULL) {
    memcpy(result, 
	   "address unknown\0",
	   strlen("address unknown\0"+1));
    return;
  }
  sprintf((char*)result,
	  "%3d.%3d.%3d.%3d:%5d",
	  ((unsigned char *)&conInfo->senderIP)[0],
	  ((unsigned char *)&conInfo->senderIP)[1],
	  ((unsigned char *)&conInfo->senderIP)[2],
	  ((unsigned char *)&conInfo->senderIP)[3],
	  conInfo->senderPort);
}

/**
 * Print the connection address (IP) encoded in the given
 * integer.
 * @param theIP the encoded IP
 **/
void printIP(int theIp,
	     IPString * result) {
  int size;
  size = sprintf((char*)result, 
		 "%3d.%3d.%3d.%3d",
		 ((unsigned char *)&theIp)[0],
		 ((unsigned char *)&theIp)[1],
		 ((unsigned char *)&theIp)[2],
		 ((unsigned char *)&theIp)[3]); 
}

void printSessionKey(SESSIONKEY * key,
		     SKEYString * result) {
  int i;
  int j = 0;
  for (i=0;i<SESSIONKEY_LEN;i++)
    j+=sprintf(&result->skey[j],"%3d- ",key->key[i]);
}

void dumpPacket(unsigned char * message,
		BLOCK_LENGTH len) {
   int total_len;
   int subpart_len;
   int pos;
   unsigned short packet_code;
   MessagePart * part;
   HELO_Body * hbody;
   QUERY_Message * qmsg;
   ACK_Message * amsg;
   SKEY_Message * skeymsg;
   HexName temp_str;

   subpart_len = -1;

   total_len = len; 
   pos = 0;

   print("!!! BEGIN PACKET - Total length: %d !!!\n", total_len);
   while (pos < total_len) {
      part = (MessagePart*) &message[pos];
      subpart_len = part->size;
      if (pos + subpart_len + BLOCK_LENGTH_SIZE > total_len) {
	print("Invalid packet format (%d - %d) - dropping remaining parts (%d).\n",
	      pos, subpart_len, total_len);	
      }
      packet_code = part->requestType;
     
      print("Packet type:\t");

      switch (packet_code) {
      case GNET_PROTO_HELO:
	print("HELO\t Length:\t %u Pos: %u\n", subpart_len, pos);
	hbody = &(((HELO_Message*)(part))->body);
	print("\n");
	break;
      case GNET_PROTO_SKEY:
	skeymsg = (SKEY_Message*) part;
	print("SKEY\t Length:\t%u Pos: %u\n", subpart_len, pos);
	break;
       case GNET_PROTO_QUERY:
	print("QUERY\t Length:\t%d Pos: %u\n", subpart_len, pos);
	qmsg = (QUERY_Message*)part; 
	print("\tQUERY:\tPriority:\t%d\t",qmsg->priority);
	print("TTL:\t%d\n",qmsg->ttl);
	hash2hex(&(qmsg->returnTo.hashPubKey),
		 &temp_str);
	print("\tQUERY:\tReturnTo:\t%s\t\n", &temp_str);
	hash2hex(&(qmsg->query), &temp_str);
	print("\tQUERY:\tQuery:\t\t%s\t\n", &temp_str);
	break;
      case GNET_PROTO_3QUERY:
	print("3QUERY\t Length:\t%d Pos: %u\n", subpart_len, pos);
	qmsg = (QUERY_Message*)part; 
	print("\t3QUERY:\tPriority:\t%d\t",qmsg->priority);
	print("TTL:\t%d\n",qmsg->ttl);
	hash2hex(&(qmsg->returnTo.hashPubKey), &temp_str);
	print("\t3QUERY:\tReturnTo:\t%s\t\n", temp_str);
	hash2hex(&(qmsg->query), &temp_str);
	print("\t3QUERY:\tQuery:\t\t%s\t\n", temp_str);
	break;
      case GNET_PROTO_CONTENT:
	print("CONTENT\t Length:\t%d Pos: %d\n", subpart_len, pos);
	break;
      case GNET_PROTO_SKEY_ACK:
	print("SKEY_ACK\t Length:\t%d (expected: %d)\n\tPos: %d\n", 
	      subpart_len, 
	      sizeof(ACK_Message)-BLOCK_LENGTH_SIZE,
	      pos);
	amsg = (ACK_Message*)part;
	hash2hex(&(amsg->hash), &temp_str);
	print("\tSKEY_ACK:\tHash:\t%s\t\t\n", &temp_str);
	assert(subpart_len == sizeof(ACK_Message)-BLOCK_LENGTH_SIZE,
	       "SKEY_ACK had unexpected size!");	       
	break;
      default: 
	print("*** UNKNOWN ***\t Length:\t%d Pos: %d\n", subpart_len, pos);
      }

      print("\n");
      pos += BLOCK_LENGTH_SIZE + part->size;
   }
   assert(pos == total_len,
	  "packet parts larger than packet analyzed!");
   print("---- END PACKET - Total length: %d ---\n", total_len);
   
}

/**
 * Assert something.
 **/
void assert(int condition,
	    char * message) {
#ifdef CHECK_ASSERTS
  if (!condition)
    print("ASSERT FAILED: %s\n",message);
#endif
}


/* end of debugging.c */
