/**
 * This module keeps track of host-identity - address (IP) translation.
 * Basically a mapping HostAddress <==> HostIdentity <==> Public Key
 * with signature and expiration date for the HostAddress.
 * <p>
 * This file basically is all about the HELO_Message.
 * @file primitives/identity.c
 * @author Christian Grothoff
 * @author Tzvetan Horozov
 **/

#include "identity.h"
#include "configuration.h"

/*for getHostAddress*/
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

HostAddress * myAddress;
DirName * networkIdDirectory;

/** 
 * Initialize identity module. Requires configuration.
 **/
void initAddress() {
  networkIdDirectory 
    = (DirName*) expandFileName((FileName)getNetworkIdDirectory());  
  myAddress = getAddress();  
}

/**
 * Bind a host identity (IP+port) to a hostId.
 * @param msg the verified (!) HELO message
 **/
void bindAddress(HELO_Message * msg) {
  HexName fil;
  FileName fn;
#ifdef PRINT_HELO
  HexName hostName;
  IPPortString ipPort;
#endif
  if (verifyHELO(msg) == 0) {
    hash2hex(&msg->body.senderIdentity.hashPubKey,
	     &fil);
    fn = (FileName) malloc(strlen((char*)networkIdDirectory) + 
			   sizeof(HexName) + 1);
    buildFileName(networkIdDirectory,
		  &fil,
		  fn);
    writeFile(fn,
	      &msg->body, 
	      sizeof(HELO_Body));
    free(fn);
  }
#ifdef PRINT_WARNINGS
  else
    print("bindAddress: HELO verification failed!\n");
#endif
#ifdef PRINT_HELO
  hash2hex(&msg->body.senderIdentity.hashPubKey,
	   &hostName);
  printAddress(&msg->body.sender, 
	       &ipPort);
  print("HELO: Bound identity %s to %s\n",
	&hostName, &ipPort);
#endif 
}

/**
 * Create the signed HELO message for this host.
 * @param msg the address where to write the HELO message to
 **/
void createHELO(HELO_Message * msg) {
  PublicKey * myKey;
  
 /* allocate memory for HELO */
  myKey = getPublicHostkey();
  /* write signed header */
  time(&msg->body.expires);
  msg->body.expires += 60 * 60 * getHeloExpiration(); 
  getHostIdentity(myKey, 
		  &msg->body.senderIdentity);
  memcpy(&msg->body.sender,
	 myAddress,
	 sizeof(HostAddress));

  /* sign sender, senderIdentity and expires */
  signData(&msg->body.sender,
	   sizeof(HostAddress)+sizeof(HostIdentity)+sizeof(time_t),
	   &msg->body.signature);  
  /* write unsigned header */
  msg->header.size = sizeof(HELO_Message)-BLOCK_LENGTH_SIZE;
  msg->header.requestType = GNET_PROTO_HELO;
  /* add public key after signature */
  memcpy(&msg->body.publicKey,
	 myKey,
	 sizeof(PublicKey));
  /* check created HELO */
  if (verifyHELO(msg) != 0) 
    errexit("Oops, HELO creation failed!\nExiting...\n");
  /* end self-check */
}

/**
 * Obtain the public key of a known host.
 * @param hostId the host id
 * @param result where to store the result
 * @returns 0 on failure, 1 on success
 **/
int identity2Key(HostIdentity *  hostId,
		 PublicKey * result) {  
  HexName fil;
  FileName fn;
  HELO_Body buffer;
  
  hash2hex(&hostId->hashPubKey,
	   &fil);
  fn = (FileName) malloc(strlen((char*)networkIdDirectory) + 
			 sizeof(HexName) + 1);
  buildFileName(networkIdDirectory,
		&fil,
		fn);
  if (!readFile(fn, sizeof(HELO_Body), &buffer)) {
    free(fn);
    return 0;  
  }
  free(fn);
  memcpy(result,
  	 &buffer.publicKey,
  	 sizeof(PublicKey));	   
  return 1;
}

/**
 * Resolve identity (find matching connection)
 * @param hostId the host id
 * @param result the connection information that was bound to that identity or NULL if no information is available
 * @return 0 on failure, 1 on success
 **/
int getConnectionInformation(HostIdentity *  hostId,
			     HostAddress * result) {
  HexName fil;
  FileName fn;
  HELO_Body buffer;
  
  hash2hex(&hostId->hashPubKey,
	   &fil);
  fn = (FileName) malloc(strlen((char*)networkIdDirectory) + 
			 sizeof(HexName) + 1);
  buildFileName(networkIdDirectory,
		&fil,
		fn);

  if (!readFile(fn, 
		sizeof(HELO_Body), 
		&buffer)) {
#ifdef PRINT_WARNINGS
    print("No connection information for host available (%s)!\n",
	  fn);
#endif
    free(fn);
    return 0;
  }
  free(fn);
  memcpy(result, &buffer.sender,
	 sizeof(HostAddress));
  return 1;
}

/**
 * Obtain the identity information for the current node
 * (connection information), conInfo.
 **/
HostAddress * getAddressFromHostname() {
  char * hostname = getenv("HOSTNAME");
  struct hostent * ip = gethostbyname(hostname);
  HostAddress * identity;
  if (ip == NULL) {
#ifdef PRINT_WARNINGS
    print("Couldn't find server '%s' (h_errno=%d)",
	  hostname, h_errno);
#endif
    return NULL;
  }
  assert(ip->h_addrtype == AF_INET,
    "getAddress: h_addrtype is not AF_INET!?");
  identity = (HostAddress*) malloc(sizeof(HostAddress));
  identity->senderPort = getGnetPort();
  memcpy(&identity->senderIP,
	 &((struct in_addr*)ip->h_addr_list[0])->s_addr,
	 sizeof(struct in_addr));
#ifdef PRINT_IDENTITY
  print("IDENTITY: getAddress (old): my address is %s\n",
	printIP(*(int*)&identity->senderIP));
#endif
  return identity;
}

/* new version - gets the IP info from the kernel */
HostAddress * getAddress(){
  IPAddress ipString = getIP();
  struct in_addr sockInfo;
  HostAddress * identity;
#ifdef PRINT_IDENTITY
  IPString ipStringDbg;
#endif
  if (ipString == NULL)
    return getAddressFromHostname();
  else
    inet_aton(ipString,&sockInfo);
  
  identity=(HostAddress*) malloc(sizeof(HostAddress));
  identity->senderPort = getGnetPort();
  memcpy(&identity->senderIP,((struct in_addr*)&sockInfo),
	 sizeof(struct in_addr));
#ifdef PRINT_IDENTITY
  printIP((*(int*)&identity->senderIP),
	  &ipStringDbg);
  print("UDP: Identity address (new): %s\n",
	&ipStringDbg);
#endif
  return identity;
}

/**
 * Send a packet to a given host. This method does the actual transfer.
 * @param hostId the id of the target host
 * @param size the size of the content
 * @param content the actual data (after the header). Encryption is
 *        NOT performed by this method! The header (MessagePack) is
 *        added, the content should consist of all message-parts.
 * @param encryptedFlag is the data encrypted (subprotocol), 
 *        or-ed with the size of the packet in the header
 * @param crc is the checksum of the unencrypted data (minus the header)
 **/
void sendToHost(HostIdentity * hostId,
		BLOCK_LENGTH size,
		void * content,
		int encryptedFlag,
                int crc) {
  MessagePack * pack;
  int sock;
  HostAddress conInfo;
  struct sockaddr_in sin;	/* an Internet endpoint address	*/
#ifdef PRINT_WARNINGS
  HexName hostName;
#endif
#ifdef PRINT_UDP
  IPString ipString;
#endif
  pack = (MessagePack *) malloc(sizeof(MessagePack) + size);
  memcpy(&pack->parts,
	 content,
	 size);
  pack->size = (size+sizeof(MessagePack)-BLOCK_LENGTH_SIZE) | encryptedFlag;
  memcpy(&pack->sender,
	 myAddress,
	 sizeof(HostAddress));
  getHostIdentity(getPublicHostkey(),
		  &pack->senderIdentity);
  pack->checksum = crc; 
  if (!getConnectionInformation(hostId, &conInfo)) {
#ifdef PRINT_WARNINGS
    hash2hex(&hostId->hashPubKey,
	     &hostName);
    print("Could not get connection information for host %s! Message not sent!\n",
	  &hostName);
#endif
    free(pack);
    return; /* could not send: no con info!*/
  }
  /* finally: UDP send! */  
  memset(&sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_port = htons(conInfo.senderPort);
  memcpy(&sin.sin_addr, &conInfo.senderIP,
	 sizeof(struct in_addr));
  sock = socket(PF_INET, SOCK_DGRAM, 17); /* 17=udp */
  if (connect(sock, (struct sockaddr *)&sin, sizeof(sin)) == 0) {
#ifdef PRINT_UDP
    printIP(*(int*)&sin.sin_addr,
	    &ipString);
    print("UDP: Sending %d bytes to %s:%d.\n",
	  size+sizeof(MessagePack), 
	  &ipString,
	  conInfo.senderPort);
#endif
    send(sock, pack, size+sizeof(MessagePack),0);
#ifdef PRINT_UDP_ALL
    print("UDP: Done with send.\n");
#endif
  }
  close(sock);
  free(pack);
}

/**
 * Verify that a HELO message is still valid. 
 * @returns 1 if the message is invalid, 0 if it is ok.len(&((PlainData)msg)[sizeof(HELO_Message)+siglen])
 **/
int verifyHELO(HELO_Message * msg) {
  int size;
  HashCode160 hashValue;

  /* check that this can actually be a HELO message */
  size = msg->header.size;
if (size != sizeof(HELO_Message)-BLOCK_LENGTH_SIZE) {
#ifdef PRINT_WARNINGS
    print("Invalid HELO (invalid size, %d bytes)!\n",
	  size);
#endif
    return 1;
  }

  /* check that the HELO has not expired */
  if (msg->body.expires < time(NULL)) {
#ifdef PRINT_WARNINGS
    print("Expired HELO-message received. Dropped!\n");
#endif
    return 1;
  }  
  /* check that the signature is valid */
  if (0 == verifySig(&msg->body.sender,
		     sizeof(HostAddress)+sizeof(HostIdentity)+sizeof(time_t),
		     &msg->body.signature,
		     &msg->body.publicKey)) {
#ifdef PRINT_WARNINGS
    print("Invalid signature for HELO!\n");
#endif
    return 1;
  }  

  /* check that the senderIdentity is really the hash of the public key */
  hash(&msg->body.publicKey,
       sizeof(PublicKey),
       &hashValue);
  if (memcmp(&hashValue,
	     &msg->body.senderIdentity,
	     sizeof(HostIdentity)) != 0) {
#ifdef PRINT_WARNINGS
    print("HELO Message sender identity does not match public key supplied.\n");
#endif
    return 1;
  }
  return 0;
}


/* end of identity.c */
