/**
 * This module is responsible to lookup queries.
 * @file gnet/primitive/lookup.c
 * @author Christian Grothoff
 **/

#include "lookup.h"
#include <dirent.h>

/* ********************* GLOBALS ***************** */

/* handle of the database-file, never closed yet */
int LOOKUP_handle_;

/* semaphore for synced r/w access to the file */
Semaphore * LOOKUP_fileSem_;

/* names of known files */
FileName * LOOKUP_fileNames_;

/* number of files that are directly shared */
unsigned int LOOKUP_fileCount_;

/* number of files that are in data/content */
unsigned int LOOKUP_filesStored_;

/* what is the priority of the least important content
   that we are storing */
int LOOKUP_leastImportance_;

/* the hashcode of the least important entry */
HashCode160 LOOKUP_lip_;

/* for removing content files... */
extern DirName * QUERY_contentDirectory_;

/* ************* internal Methods **************** */

/**
 * Cron-job that makes sure the lookup-module is fine
 * by synchronizing the in-memory database with the drive.
 **/
void cronSyncLookup();

/**
 * Cron-job that decreases the importance-level of all
 * files by 1. Runs 'not very often'.
 **/
void cronReduceImportance();

/* scan the list of on-demand shared files to
   initialize LOOUP_fileNames_ */
void scanDatabaseList();

/**
 * Given an entry in the cache, obtain the matching
 * message(s). Send them to the given host.
 **/ 
void resolveEntry(ContentIndex * entry,
		  HostIdentity * replyTo);

/**
 * Lookup the given Hash in the Hashtable-cache structure
 * (load blocks as needed).
 * @returns 0 if not found, 1 if found
 **/
int findEntry(HashCode160 * query,
	      ContentIndex * entry);

/**
 * Write the given entry to the database.
 **/
void writeEntry(ContentIndex * entry);

/**
 * Delete an entry and the corresponding file on the
 * drive.
 **/
void deleteEntry(HashCode160 * entry);

/**
 * Find the last important entry and
 * set the LOOKUP_lip_ and LOOKUP_leastImportance_
 * variables accordingly.
 **/
void findLeastImportant();

/* ******************* CODE... ****************** */

/**
 * Initialize the lookup-module.
 **/
void initLookup() {  
  FileName fileName;
  FileName dirName;
  ContentIndex entry;
  DIR * dinfo;
  struct dirent *finfo;
  struct stat istat;
    
  print("Indexing data/content database... ");
  LOOKUP_leastImportance_ = 0;
  initLookup_collision();
  /* scan lookup-file index */
  scanDatabaseList();

  fileName = getDatabaseFile();
  fileName = expandFileName(fileName);
  LOOKUP_handle_ = open(fileName, O_CREAT|O_RDWR, S_IRUSR|S_IWUSR);  
  free(fileName);

  if (LOOKUP_handle_ < 0)
    errexit("Could not open/create database!\n");
  LOOKUP_fileSem_ = new_semaphore(1);

  /* scan directory data/content and add entries to database 
     if not already present */
  LOOKUP_filesStored_ = 0;

  dirName = expandFileName((FileName)getContentDirectory());
  stat((char*)dirName,&istat);
  if (!S_ISDIR(istat.st_mode)) {
#ifdef PRINT_WARNINGS
    print("contentDirectory must be invoked on a directory (%s)!\n",
	  dirName);
#endif
    free(dirName);
    return;
  }
  errno = 0;
  dinfo = opendir((char*)dirName);
  if ((errno == EACCES)||(dinfo == NULL)) {
    print("initLookup: access denied (%s)\n",dirName);
    free(dirName);
    return;
  }
  while ((finfo = readdir(dinfo)) != NULL)
    if (strlen(finfo->d_name) == sizeof(HashCode160)*2) {	
      hex2hash((HexName*)finfo->d_name,
	       &entry.doubleHash);
      entry.importance = 0;
      entry.fileNameIndex = -1; /* directly available in data/content */
      entry.fileOffset = -1; 
      addEntry(&entry);
      LOOKUP_filesStored_++; /* one more file */
    }
  closedir(dinfo);
  print("scanning existing database... ");
  findLeastImportant();
  /* schedule cron job */
  addCronJob(&cronSyncLookup,
	     5*60, 5*60); /* flush every 5 minutes */
  /* schedule cron job */
  addCronJob(&cronReduceImportance,
	     6*60*60, 12*60*60); /* decrease every 12h, but first time after 6h */
  print("Database initialization complete.\n");
}

/* scan the list of on-demand shared files to
   initialize LOOUP_fileNames_ */
void scanDatabaseList() {
  FileName fil;
  FILE * handle;
  int result;
  unsigned char line[65536];

  LOOKUP_fileCount_ = 0;
  fil = getSharedFileList();
  fil = expandFileName(fil);   
  handle = fopen(fil, "a+");
  fseek(handle,0,SEEK_SET);
  if (handle == NULL) {
#ifdef PRINT_WARNINGS
    print("LOOKUP: List %s of directly shared filenames not available!\n",
	  fil);
#endif    
    return;
  }
  result = 1;
  LOOKUP_fileCount_ = -1;
  while (result != EOF) {
    result = fscanf(handle, "%65535s\n", line);
    LOOKUP_fileCount_++;
  }
  if (LOOKUP_fileCount_ == 0) {
    free(fil);
    fclose(handle);
    return;
  }
  fseek(handle, 0, SEEK_SET);
  LOOKUP_fileNames_ = malloc(LOOKUP_fileCount_ * sizeof(FileName));
  LOOKUP_fileCount_ = 0;
  result = 1;
  while (result != EOF) {
    result = fscanf(handle, "%65535s\n", line);
    if (result != EOF) {
      LOOKUP_fileNames_[LOOKUP_fileCount_] = malloc(strlen(line)+1);
      memcpy(LOOKUP_fileNames_[LOOKUP_fileCount_],
	     line,
	     strlen(line)+1);
      LOOKUP_fileCount_++;
    }
  }
  free(fil);
  fclose(handle);  
}

/**
 * Add a name to the list of filenames.
 **/
int appendFilename(char * filename) {
  FileName fil;
  FILE * handle;
  int result;
  unsigned char line[65536];

  LOOKUP_fileCount_ = 0;
  fil = getSharedFileList();
  fil = expandFileName(fil);   
  handle = fopen(fil, "r+");
  free(fil);
  filename = expandFileName(filename);   
  if (handle == NULL) 
    errexit("LOOKUP: List %s of directly shared filenames not available!\n",
	    filename);
  result = 1;
  LOOKUP_fileCount_ = 0;
  while (result != EOF) {
    result = fscanf(handle, "%65535s\n", line);
    LOOKUP_fileCount_++;
  }
  fprintf(handle,"%s\n",filename);
  fclose(handle);  
  free(filename);  
  return LOOKUP_fileCount_; /* return index */
}

/**
 * Is space available on the drive? Returns the number of
 * blocks (1k) available. If there would be no space except
 * if less important content is discarded, this method 
 * *deletes* 1k of the less important content and returns 
 * 1. If there is only more important content, the
 * method returns 0.
 * @param importance how important is the content?
 **/
int spaceAvailable(int importance) {  
  if (LOOKUP_filesStored_ > getDiskQuota() * 1024) {
    if (importance <= LOOKUP_leastImportance_)
      return 0; /* no */
    /* free lip entry */
    deleteEntry(&LOOKUP_lip_);
    findLeastImportant();
    return 1;
  } else {
    LOOKUP_filesStored_++; /* it is safe to assume that there will be one more file very soon... */
    return getDiskQuota() * 1024 - LOOKUP_filesStored_; /* this many blocks are free */
  }     
}

/**
 * Cron-job that makes sure the lookup-module is fine
 * by synchronizing the in-memory database with the drive.
 **/
void cronSyncLookup() {
  assert(fdatasync(LOOKUP_handle_) == 0,
	 "fdatasync failed on database!");
}

/**
 * Cron-job that decreases the importance-level of all
 * files by 1. Runs 'not very often'.
 **/
void cronReduceImportance() {
  ContentIndex entry;
  int i = -1;  
  int j = sizeof(ContentIndex);
  semaphore_down(LOOKUP_fileSem_);
  lseek(LOOKUP_handle_, 0, SEEK_SET);
  while (j == sizeof(ContentIndex)) {
    i++;
    j = read(LOOKUP_handle_, &entry, sizeof(ContentIndex));
    if (j == sizeof(ContentIndex)) {
      if (entry.fileNameIndex >= 0)
	continue; /* on demand encoding or free */
      if (entry.importance > 0) {
	entry.importance--;
	lseek(LOOKUP_handle_, i*sizeof(ContentIndex), SEEK_SET);
	write(LOOKUP_handle_, &entry, sizeof(ContentIndex));     
      }
      if (entry.importance < LOOKUP_leastImportance_) {
	LOOKUP_leastImportance_ = entry.importance;
	memcpy(&LOOKUP_lip_,
	       &entry.doubleHash,
	       sizeof(HashCode160));
      }
    }    
  }
  semaphore_up(LOOKUP_fileSem_);
}

/**
 * Delete an entry and the corresponding file on the
 * drive.
 * @param entry the double-hash of the item to remove (=filename)
 **/
void deleteEntry(HashCode160 * entry) {
  int noll = 0;
  int index;
  HashCode160 tripleHash;
  FileName fn;
  HexName fil;
  int i;

  /* remove entry from database */
  hash(&entry,
       sizeof(HashCode160),
       &tripleHash);
  index = tripleHash.a & (CACHE_HASH_TABLE_SIZE-1);
  index *= sizeof(ContentIndex);
  semaphore_down(LOOKUP_fileSem_);
  lseek(LOOKUP_handle_, index, SEEK_SET);
  /* fill database with zeros */
  for (i=0;i<(sizeof(ContentIndex)>>2);i++)
    write(LOOKUP_handle_, &noll, sizeof(int));
  semaphore_up(LOOKUP_fileSem_);

  /* remove the file */
  hash2hex(entry, &fil); 
  fn = malloc(strlen((char*)QUERY_contentDirectory_)+
	      sizeof(HexName)+1);
  buildFileName(QUERY_contentDirectory_, 
		&fil,
		fn);
  unlink(fn);
  free(fn);
}

/**
 * Find the last important entry and
 * set the LOOKUP_lip_ and LOOKUP_leastImportance_
 * variables accordingly.
 **/
void findLeastImportant() {
  ContentIndex entry;
  int j = sizeof(ContentIndex);
  semaphore_down(LOOKUP_fileSem_);
  LOOKUP_leastImportance_ = 0x7FFFFFFF; /* max */
  lseek(LOOKUP_handle_, 0, SEEK_SET);
  while (j == sizeof(ContentIndex)) {
    j = read(LOOKUP_handle_, &entry, sizeof(ContentIndex));
    if (j == sizeof(ContentIndex)) {
      if (entry.fileNameIndex >= 0)
	continue; /* on demand encoding or free */
      if (entry.importance < LOOKUP_leastImportance_) {
	LOOKUP_leastImportance_ = entry.importance;
	memcpy(&LOOKUP_lip_,
	       &entry.doubleHash,
	       sizeof(HashCode160));
      }
    }    
  }
  semaphore_up(LOOKUP_fileSem_);
}

/**
 * Lookup the file on the local machine. If found,
 * send it back.
 * @param query the triple-hash of the content
 * @param replyTo the host that was asking for it
 * @param qp the polciy (for us: priority)
 * @returns 1 if found, 0 if not.
 **/
int tripleHashLookup(HashCode160 * query,
		     HostIdentity * replyTo,
		     QUERY_POLICY qp) {
  ContentIndex ci;

  if (0 == findEntry(query,&ci))
    return 0;
  ci.importance += qp; /* increase importance */
  writeEntry(&ci); /* write to disk! */
  resolveEntry(&ci, replyTo); /* send reply */
  return 1; /* report success */
}

/**
 * Given an entry in the cache, obtain the matching
 * message(s). Send them to the given host.
 **/ 
void resolveEntry(ContentIndex * entry,
		  HostIdentity * replyTo) {
  CONTENT_Block result;
  CONTENT_Block crypt;
  FileName fn;
  int fileHandle;
  ssize_t blen;
  HashCode160 hc;

  /* data/content mechanism */
  if (entry->fileNameIndex == -1) {
    localLookup(&entry->doubleHash,
		replyTo);
    return;
  }
  if (entry->fileNameIndex > LOOKUP_fileCount_) {
#ifdef PRINT_WARNINGS
    errexit("Database inconsistent! (index points to invalid offset %d (>= %d)\n",
	    entry->fileNameIndex, LOOKUP_fileCount_);
    return;
#endif
  }
  /* on-demand encoding mechanism */
  fn = LOOKUP_fileNames_[entry->fileNameIndex-1];
  fileHandle = open(fn, O_EXCL, S_IRUSR);
  if (fileHandle == -1) {
#ifdef PRINT_WARNINGS
    print("Could not open file %s.\n",
	  fn);    
#endif
    return;
  }
  lseek(fileHandle, entry->fileOffset,SEEK_SET);
  blen = read(fileHandle, 
	      &result,
	      sizeof(CONTENT_Block));
  close(fileHandle);
  hash(&result, 
       blen, 
       &hc);
  if (encryptContent(&result,
		     &hc,
		     &crypt))
    sendReply(replyTo, 
	      &crypt, 
	      &entry->doubleHash);  
}

/**
 * Lookup the given Hash in the Hashtable-cache structure
 * (load blocks as needed).
 * @returns 0 if not found, 1 if found
 **/
int findEntry(HashCode160 * query,
	      ContentIndex * entry) {
  unsigned int index;
  HashCode160 tripleHash;

  index = query->a & (CACHE_HASH_TABLE_SIZE-1);
  index *= sizeof(ContentIndex);
  semaphore_down(LOOKUP_fileSem_);
  lseek(LOOKUP_handle_, index, SEEK_SET);
  read(LOOKUP_handle_, entry, sizeof(ContentIndex));
  semaphore_up(LOOKUP_fileSem_);
  hash(&entry->doubleHash,
       sizeof(HashCode160),
       &tripleHash);
  if (entry->fileNameIndex == 0) 
    return 0; /* no collision, nothing there */
  if (equalsHashCode160(query, &tripleHash)) 
    return 1;
  else 
    return findCollisionEntry(query, entry);  
}

/**
 * Write the given entry to the database. 
 **/
void addEntry(ContentIndex * entry) {
  unsigned int index;
  int size;

  HashCode160 tripleHash;
  ContentIndex onDisk;

  hash(&entry->doubleHash,
       sizeof(HashCode160),
       &tripleHash);

  index = tripleHash.a & (CACHE_HASH_TABLE_SIZE-1);
  index *= sizeof(ContentIndex);
  semaphore_down(LOOKUP_fileSem_);
  lseek(LOOKUP_handle_, index, SEEK_SET);
  size = read(LOOKUP_handle_, &onDisk, sizeof(ContentIndex));
  semaphore_up(LOOKUP_fileSem_);
  if ((size == 0) || (onDisk.fileNameIndex == 0)) {
    writeEntry(entry);
    return; /* no collision, nothing there, just write */
  }
  if (equalsHashCode160(&onDisk.doubleHash, &entry->doubleHash)) {
    writeEntry(entry); /* update */
    return;
  }
  /* got to collision DB */
  addCollisionEntry(entry);
  return;  
}

/**
 * Write the given entry to the database.
 **/
void writeEntry(ContentIndex * entry) {
  unsigned int index;
  HashCode160 tripleHash;
  hash(&entry->doubleHash,
       sizeof(HashCode160),
       &tripleHash);
  
  index = tripleHash.a & (CACHE_HASH_TABLE_SIZE-1);
  index *= sizeof(ContentIndex);
  semaphore_down(LOOKUP_fileSem_);
  lseek(LOOKUP_handle_, index, SEEK_SET);
  write(LOOKUP_handle_, entry, sizeof(ContentIndex));
  semaphore_up(LOOKUP_fileSem_);
}


/* end of lookup.c */
