/**
 * This module is responsible to handle queries.
 * @file gnet/primitive/query.c
 * @author Christian Grothoff
 **/

#include "query.h"

DirName * QUERY_contentDirectory_;

void initQuery() {
  QUERY_contentDirectory_
    = (DirName*) expandFileName((FileName)getContentDirectory());
}

/**
 * Check if the file the query is for is on the
 * local drive. If yes, send the reply and return 1,
 * otherwise return 0.
 * @param query the filename to look for
 * @param replyTo the host to send replies to
 * @param qp the priority to apply for the reply
 **/
int localLookup(HashCode160 * query,
		HostIdentity * replyTo) {
  CONTENT_Block buf[MAX_FSIZE];
  HexName fil;
  FileName fn;
  int i,j;
 
#ifdef PRINT_QUERY
  HexName hex;
  hash2hex(query, &hex);

  print("QUERY: local lookup for %s.\n",
	&hex);
#endif

  hash2hex(query, &fil); 
  fn = malloc(strlen((char*)QUERY_contentDirectory_)+
	      sizeof(HexName)+1);
  buildFileName(QUERY_contentDirectory_, 
		&fil,
		fn);
  j = readFile(fn, MAX_FSIZE * CONTENT_SIZE, buf); 
  if (j <= 0) {
#ifdef PRINT_QUERY
    print("QUERY: local lookup: %s not found!\n",fn);
#endif
    free(fn);
    return 0;
  }
  /* data found on local node! */
  if ((j % CONTENT_SIZE) != 0) {
#ifdef PRINT_WARNINGS
    print("Bad content file: %s\n",fn);
#endif
    free(fn);
    return 0;
  }
#ifdef PRINT_QUERY
  print("QUERY: local lookup succeded. Sending reply (%s)!\n",
	fn);
#endif
  free(fn);

  /* send multiple replies if applicable */
  j=j/CONTENT_SIZE;
  for (i=0;i<j;i++) 
    sendReply(replyTo, 
	      &buf[i], 
	      query);  
#ifdef PRINT_QUERY_ALL
  print("QUERY: local lookup done\n");
#endif
  return 1;
}

/**
 * Send a reply.
 * @param replyTo to whom to send the reply
 * @param content the data to send (1k)
 **/
void sendReply(HostIdentity * replyTo,
	       CONTENT_Block * content,
	       HashCode160 * hash) {
  CONTENT_Message reply;
#ifdef PRINT_QUERY
  HexName fil;
  HexName query;

  hash2hex(&replyTo->hashPubKey,
	   &fil);
  hash2hex(hash, &query);
  print("QUERY: unicasting reply %s to %s\n",
	&query,
	&fil);
#endif

  reply.header.size = sizeof(CONTENT_Message)-BLOCK_LENGTH_SIZE;
#ifdef PRINT_QUERY_ALL
  print("QUERY: sending reply of size %d\n",
	sizeof(CONTENT_Message));
#endif
  reply.header.requestType = GNET_PROTO_CONTENT;
  memcpy(&reply.reply.content,
	 content,
	 CONTENT_SIZE);
  memcpy(&reply.reply.hash,
	 hash,
	 sizeof(HashCode160));
  unicast(&reply.header, 
	  replyTo, 
	  0);  
  /* priority 0 may be inadequate, FIX ME, too! */
#ifdef PRINT_QUERY_ALL
  print("QUERY: unicasting reply done\n");
#endif
}
 

/**
 * compare two hashcodes.
 **/
int equalsHashCode160(HashCode160 * a, 
		      HashCode160 * b) {
  return 
    (a->a == b->a) &&
    (a->b == b->b) &&
    (a->c == b->c) &&
    (a->d == b->d) &&
    (a->e == b->e);
}

/**
 * Execute the query. This method does NOT check the
 * QUERY Policy, it is used either after the policy
 * was checked or if the request is trusted (read:
 * comes from our GProxy).
 **/
void execQuery(QUERY_POLICY qp, QUERY_Message * msg) {
  if ((qp & QUERY_ANSWER) > 0) {
    if (localLookup(&msg->query,
		    &msg->returnTo))
      return; 
  } /* end of "found" */

  if (msg->ttl <= 0)
    return; /* timeout */
 
  if (isPending(msg)) {
    indirectingQuery(msg); 
    return; /* we're already searching for it */
  }
  if ((qp & QUERY_FORWARD) > 0) {
    msg->priority = qp & QUERY_PRIORITY_BITMASK; /* reduce priority to assigned priority */
    /* forward query */
    if ((qp & QUERY_INDIRECT) > 0) { 
#ifdef PRINT_QUERY
      print("QUERY: indirecting query (TTL %d)\n", msg->ttl);
#endif
      /* forward query & replace returnTo with our identity, 
	 also keep track of the original returnTo for the case
	 we get a reply! */
      indirectingQuery(msg); 
      /* set reply-to to myself */
      getHostIdentity(getPublicHostkey(),
		      &msg->returnTo);
    } else 
      msg->priority = 0; /* we're not even indirecting, so we can't earn credit, so we don't care */
    msg->ttl -= TTL_DECREMENT;
    if (msg->ttl <= 0)
      msg->ttl = 0;
    else
      forwardQuery(msg);
  }
}

/**
 * Execute the triple-hash query. This method does NOT check the
 * QUERY Policy, it is used either after the policy
 * was checked or if the request is trusted (read:
 * comes from our GProxy).
 **/
void exec3Query(QUERY_POLICY qp, QUERY_Message * msg) {
  if ((qp & QUERY_ANSWER) > 0) 
    if (tripleHashLookup(&msg->query,
			 &msg->returnTo,
			 qp))
      return;  /* end of "found" */
  if (msg->ttl <= 0)
    return; /* timeout */
  
  /* indirection is yet not implemented for
     triple-hashed scheme */
  if (isPending(msg)) {
    indirectingQuery(msg); 
    return; 
  }
  if ((qp & QUERY_FORWARD) > 0) {
    msg->priority = qp & QUERY_PRIORITY_BITMASK; /* reduce priority to assigned priority */
    /* forward query */
    if ((qp & QUERY_INDIRECT) > 0) { 
#ifdef PRINT_QUERY
      print("QUERY: indirecting query\n");
#endif
      /* forward query & replace returnTo with our identity, 
	 also keep track of the original returnTo for the case
	 we get a reply! */
      indirectingQuery(msg); 
      /* set reply-to to myself */
      getHostIdentity(getPublicHostkey(),
		      &msg->returnTo);
    } else 
      msg->priority = 0; /* we're not even indirecting, so we can't earn credit, so we don't care */
    msg->ttl -= TTL_DECREMENT;
    if (msg->ttl <= 0)
      msg->ttl = 0;
    else
      forwardQuery(msg);
  }
}

/* end of query.c */

