/**
 * sessionkey management
 * @author Christian Grothoff
 * @file primitives/sessionkey.c
 **/

#include "gnettypes.h"
#include "sessionkey.h"

/**
 * Force creation of a new Session key for the given host. 
 * @param hostId the identity of the other host 
 * @param sk the SESSIONKEY to use
 * @param created the timestamp to use
 * @param ret the address where to write the signed session key, first BLOCK_LENGTH_SIZE byte give length
 * @return 1 on success, 0 on failure
 **/
int makeSessionKeySigned(HostIdentity * hostId,
			 SESSIONKEY * sk,
			 time_t created,
			 SKEY_Message * ret) {
  SessionKeyEncoded enc;
  HashCode160 keyHash;
  HostIdentity myId;
  PublicKey otherPubKey;
#ifdef EXTRA_CHECKS
  HexName hostName;
#ifdef EXTRA_CHECKS_LOOPBACK_ONLY
  SessionKeyEncoded key;
  SKEYString skeyString1;
  SKEYString skeyString2;
#endif
#endif

  /* create and encrypt sessionkey */
  if (!identity2Key(hostId, &otherPubKey)) {
#ifdef PRINT_WARNINGS
    print("makeSesionKeySigned: cannot encrypt sessionkey, other host not known!\n");  
#endif 
    return 0; /* other host not known */  
  }
  enc.len = SESSIONKEY_LEN;
  memcpy(&enc.skey,
	 sk,
	 sizeof(SESSIONKEY));

  if (!encrypt(&enc,
	       sizeof(SessionKeyEncoded),
	       &otherPubKey,
	       &ret->body.key)) {
#ifdef PRINT_WARNINGS
    print("makeSesionKeySigned: encrypt failed!");
#endif
    return 0; /* encrypt failed */
  }
#ifdef EXTRA_CHECKS_LOOPBACK_ONLY
  if ((sizeof(SessionKeyEncoded) != 
       decryptData(&ret->body.key,
		   &key,
		   sizeof(SessionKeyEncoded))) ||
      (key.len != SESSIONKEY_LEN)) {
    printSessionKey(&enc.skey, &skeyString1);
    printSessionKey(&key.skey, &skeyString2);
    print("encoded skey: %s (%d)\ndecoded %s (%d)\n",
	  &skeyString1, enc.len, &skeyString1, key.len);
  }  
#endif
  /* compute hash and sign hash */
  ret->body.created = created;
  hash(&ret->body.key, 
       sizeof(RSAEncryptedData) + sizeof(time_t),
       &keyHash);
  signData(&keyHash, 
	   (BLOCK_LENGTH)sizeof(HashCode160),
	   &ret->body.signature);

  /* complete header */
  ret->header.size = sizeof(SKEY_Message) - BLOCK_LENGTH_SIZE;
  ret->header.requestType = GNET_PROTO_SKEY;
  /* verify signature/SKS */
  getHostIdentity(getPublicHostkey(),
		  &myId); 
#ifdef EXTRA_CHECKS
  hash2hex(&hostId->hashPubKey, &hostName);
  if (0 == verifySKS(&myId, ret)) {
    errexit("MakeSessionKey: generated sessionkey for host %s failed verification! exiting!\n",
	    &hostName);
  }
#endif
  return 1;
} 


/**
 * @return 0 if invalid, 1 if valid
 **/
int verifySKS(HostIdentity * hostId,
	      SKEY_Message * sks) {
  PublicKey pubKey;
  HashCode160 keyHash;
#ifdef PRINT_WARNINGS
  HexName hostName;
#endif
  
  if ( (sks == NULL) ||
       (hostId == NULL)) {
#ifdef PRINT_WARNINGS
    hash2hex(&hostId->hashPubKey, &hostName);
    print("VerifySKS: invalid sessionkey suggestion for host %s\n",
	  &hostName);
#endif
    return 0;
  }
  
  if (!identity2Key(hostId, &pubKey)) {
#ifdef PRINT_WARNINGS
    hash2hex(&hostId->hashPubKey, &hostName);
    print("verifySKS: host %s for sessionkey exchange not known\n",
	  &hostName);
#endif
    return 0;
  }
  /* verify signature */
  hash(&sks->body.key, 
       sizeof(RSAEncryptedData) + sizeof(time_t),
       &keyHash);
  if (!verifySig(&keyHash,
		 sizeof(HashCode160),
		 &sks->body.signature, 
		 &pubKey)) {
#ifdef PRINT_WARNINGS
    print("verifySKS: received SessionKey with bad signature!\n");
#endif
    return 0; /*reject!*/
  }
  return 1; /* ok */
}

/**
 * Accept a session-key that has been sent by another host.
 * The other host must be known (public key)
 * @param hostId the identity of the sender host
 * @param sessionkey the session key that was "negotiated"
 * @param sig the signature to verify the sessionkey is ok
 **/
void acceptSessionKey(HostIdentity * hostId,
		      SKEY_Message * sessionkeySigned) {
  SessionKeyEncoded key;
  ACK_Message ack;
#ifdef PRINT_SKEY_ALL
  HexName hostName;
#endif
#ifdef PRINT_WARNINGS_ALL
  HostAddress conInfo;
#ifndef PRINT_SKEY_ALL
  HexName hostName;
  IPPortString ipPort;
#endif
#endif

#ifdef PRINT_SKEY_ALL
  hash2hex(&hostId->hashPubKey,
	    &hostName);
  print("SKEY: Accepting sessionkey from host %s.\n",
	&hostName);
#endif
  if (0 == verifySKS(hostId, sessionkeySigned))
    return;  /* rejected */
#ifdef PRINT_SKEY_ALL
  print("SKEY: Key was verified.\n");
#endif

  /* prepare file */
  if ((sizeof(SessionKeyEncoded) != 
       decryptData(&sessionkeySigned->body.key,
		   &key,
		   sizeof(SessionKeyEncoded))) ||
      (key.len != SESSIONKEY_LEN)) {
#ifdef PRINT_WARNINGS_ALL
#ifndef PRINT_SKEY_ALL
    hash2hex(&hostId->hashPubKey,
	     &hostName);
#endif
    if (getConnectionInformation(hostId,
				 &conInfo)) {
      printAddress(&conInfo,
		   &ipPort);
      print("SKEY rejected from host %s (address %s): key.len = %d\n",
	    &hostName,
	    &ipPort,
	    key.len);
    } else
      print("SKEY rejected from host %s (address unknown): key.len = %d\n",
	    &hostName,
	    key.len);
#endif    
    return;
  }
  if (addHostWithSKEY(hostId, &key.skey, 
		      sessionkeySigned->body.created)) {
    return;
  }
  notifyActive(hostId);
  /* Prepare an ACK */
  makeAckMessage(&key,&ack);

  unicast(&ack.header, hostId, 0);  
  sendBufferOfHost(hostId);
}

/**
 * The following routine constructs a receipt for the 
 * session key.
 * @param sessionkey the key to use
 * @param ret address where to write the receit to
 **/
void makeAckMessage(SessionKeyEncoded * sessionKey,
		    ACK_Message * ret){
  hash(&sessionKey->skey,
       sizeof(SESSIONKEY),
       &ret->hash);
  ret->header.size = sizeof(ACK_Message)-BLOCK_LENGTH_SIZE;
  ret->header.requestType = GNET_PROTO_SKEY_ACK;  
}

/* end of sessionkey.c */
