/**
 * sessionkey - standard communication with another node.
 * encryptContent - given a data block and the key (i.e. hashcode)
 *                return the encrypted data block
 * decryptContent - given an ecrypted data block and a key, retrieve the
 *                useful information by decryption
 * @author Christian Grothoff
 * @author Ioana Patrascu
 * @file primitives/sessionkey.c
 **/

#include "sessionkey.h"
#include "storage.h"

#define INITVALUE "GNET!!!!"

/**
 * Create a new SessionKey (for Blowfish)
 **/
void makeSessionkey(SESSIONKEY * key) {
  int i;
  for (i=0;i<SESSIONKEY_LEN;i++)
    key->key[i] = rand();
}

/**
 * Encrypt a block with the public key of another
 * host that uses the same cyper.
 * @param block the block to encrypt
 * @param len the size of the block
 * @param sessionkey the key used to encrypt
 * @returns the size of the encrypted block, -1 for errors
 **/
int encryptBlock(void * block, 
		 BLOCK_LENGTH len,
		 SESSIONKEY * sessionkey,
		 BLOWFISHEncryptedData * result) {
  int outlen = 0;
  unsigned char iv[] = INITVALUE;
  EVP_CIPHER_CTX ctx;
  
  if ((block == NULL) || (sessionkey == NULL)) 
    return -1;  
  /* compute result size by adding block-length, always padded */
  EVP_EncryptInit(&ctx, EVP_bf_cfb(), sessionkey->key, iv);
  if (0 == EVP_EncryptUpdate(&ctx, 
			     &result->encoding[0], 
			     &outlen,
			     block, len)) {
    print("Sessionkey.c-encryptBlock: EVP_EncryptUpdate failed!\n");
    return -1;
  }
  len = outlen; /* save bytes written so far */
  outlen = 0;
  if (0 == EVP_EncryptFinal(&ctx, &result->encoding[0], &outlen)) {
    print("Sessionkey.c-encryptBlock: EVP_EncryptFinal failed!\n");
    return -1;
  }
  outlen += len; /* add both updates together */
  EVP_CIPHER_CTX_cleanup(&ctx);
  return outlen;
}

/**
 * Decrypt a given block with the sessionkey.
 * @param sessionkey the key used to decrypt
 * @param block the data to decrypt, encoded as returned by encrypt
 * @param result address to store the result at
 * @return -1 on failure, size of decrypted block on success
 **/
int decryptBlock(SESSIONKEY * sessionkey, 
		 BLOWFISHEncryptedData * block,
		 BLOCK_LENGTH size,
		 void * result) {
  int outlen = 0;
  unsigned char iv[] = INITVALUE;
  EVP_CIPHER_CTX ctx;

  /* use blowfish-cfb */
  EVP_DecryptInit(&ctx, EVP_bf_cfb(), sessionkey->key, iv);
  if (0 == EVP_DecryptUpdate(&ctx, result, &outlen,
			     block->encoding, size)) {
    print("Sessionkey.c-decryptBlock: EVP_DecryptUpdate failed!\n");
    return -1;
  }
  size = outlen;
  outlen = 0; 
  if (0 == EVP_DecryptFinal(&ctx, &((unsigned char*)result)[size], &outlen)) {
    print("Sessionkey.c-decryptBlock: EVP_DecryptFinal failed!\n");
    return -1;
  }
  return size+outlen;
}

/**
 * Encrypts a given data block
 *
 * @param data represents the data block
 * @param hashcode represents the key concatenated with the initial
 *        value used in the alg
 * @param result where to store the result (encrypted block)
 * @returns 1 on success, 0 on error
 **/
int encryptContent(CONTENT_Block * data,
		   HashCode160 * hashcode,
		   CONTENT_Block * result){
  unsigned char iv[BLOWFISH_BLOCK_LENGTH];            /* initial value */
  EVP_CIPHER_CTX ctx;           /* context used in the encryption alg */
  int tmplen;
  int outlen = 0;  		/* output length */

  if ((data == NULL) || (hashcode == NULL) || (result == NULL))
    return 0;
  /* get key and init value from the hash code */
  memcpy(iv, &(((char *)hashcode)[BF_KEYSIZE]), BLOWFISH_BLOCK_LENGTH/2);
  memcpy(&iv[BLOWFISH_BLOCK_LENGTH/2], &(((char *)hashcode)[BF_KEYSIZE]),
         BLOWFISH_BLOCK_LENGTH/2);
  /* encrypt with hashcode */
  EVP_EncryptInit(&ctx, EVP_bf_cfb(), (unsigned char*) hashcode, iv);
  if (0 == EVP_EncryptUpdate(&ctx, 
			     &result->content[0],
			     &outlen, 
			     &data->content[0], 
			     CONTENT_SIZE)) {
    print("encrypt.c-encryptBlock: EVP_EncryptUpdate failed!\n");
    return 0;
  }

  /* since our blocks are multiples of 64, this should be
     unnecessary. But it can't hurt either and will
     prevent bugs if we ever change to a size that is not
     a multiple of 64 bit (very unlikely :-). */
  if (0 == EVP_EncryptFinal(&ctx, 
			    &result->content[outlen], 
			    &tmplen)) {
    print("encrypt.c-encryptBlock: EVP_EncryptFinal failed!\n");
    return 0;
  }
  EVP_CIPHER_CTX_cleanup(&ctx);
  /* return encrypted block */
  return 1;
}


/* end of sessionkey.c */
